# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import os.path
from collections.abc import Mapping
from functools import cache
from typing import Any, Optional

from xcube.constants import DEFAULT_SERVER_ADDRESS, DEFAULT_SERVER_PORT
from xcube.util.jsonschema import (
    JsonArraySchema,
    JsonBooleanSchema,
    JsonIntegerSchema,
    JsonObjectSchema,
    JsonStringSchema,
)

BASE_SERVER_CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        port=JsonIntegerSchema(title="Server port.", default=DEFAULT_SERVER_PORT),
        address=JsonStringSchema(
            title="Server address.", default=DEFAULT_SERVER_ADDRESS
        ),
        base_dir=JsonStringSchema(
            title="Base directory used to resolve relative local paths."
            " Can be a local filesystem path or an absolute URL.",
        ),
        url_prefix=JsonStringSchema(
            title="Prefix to be prepended to all URL route paths."
            " Can be an absolute URL or a relative URL path.",
        ),
        reverse_url_prefix=JsonStringSchema(
            title="Prefix to be prepended to reverse URL paths"
            " returned by server responses."
            " Can be an absolute URL or a relative URL path.",
        ),
        trace_perf=JsonBooleanSchema(
            title="Output performance measures",
        ),
        static_routes=JsonArraySchema(
            title="Static content routes",
            items=JsonObjectSchema(
                properties=dict(
                    path=JsonStringSchema(title="The URL path", min_length=1),
                    dir_path=JsonStringSchema(
                        title="A local directory path", min_length=1
                    ),
                    default_filename=JsonStringSchema(
                        title="Optional default filename",
                        examples=["index.html"],
                        min_length=1,
                    ),
                    openapi_metadata=JsonObjectSchema(
                        title="Optional OpenAPI operation metadata",
                        additional_properties=True,
                    ),
                ),
                required=["path", "dir_path"],
                additional_properties=False,
            ),
        ),
        api_spec=JsonObjectSchema(
            title="API specification",
            description="selected = (includes | ALL) - (excludes | NONE)",
            properties=dict(
                includes=JsonArraySchema(JsonStringSchema(min_length=1)),
                excludes=JsonArraySchema(JsonStringSchema(min_length=1)),
            ),
            additional_properties=False,
        ),
    ),
    # We allow for other configuration settings contributed
    # by APIs. If these APIs are currently not in use,
    # validation would fail if additional_properties=False.
    additional_properties=True,
)


@cache
def normalize_base_dir(base_dir: Optional[str]) -> str:
    """Normalize the given base directory *base_dir*."""
    if base_dir is None:
        base_dir = os.path.abspath("")
    elif not is_absolute_path(base_dir):
        base_dir = os.path.abspath(base_dir)
    while base_dir != "/" and base_dir.endswith("/"):
        base_dir = base_dir[:-1]
    return base_dir


def get_base_dir(config: Mapping[str, Any]) -> str:
    """Get the normalized base directory from configuration *config*."""
    return normalize_base_dir(config.get("base_dir"))


def resolve_config_path(config: Mapping[str, Any], path: str) -> str:
    """Resolve a given relative *path* against the base directory given by
    *config*. Return *path* unchanged, if it is absolute.
    """
    if is_absolute_path(path):
        abs_path = path
    else:
        base_dir = get_base_dir(config)
        abs_path = f"{base_dir}/{path}"
    # Resolve ".." and "." in path
    if "://" in abs_path:
        scheme, host_path = abs_path.split("://", maxsplit=1)
        if "/" in host_path:
            hostname, url_path = host_path.split("/", maxsplit=1)
            url_path = _remove_path_dot_segments(url_path)
            return f"{scheme}://{hostname}/{url_path}"
        else:
            return f"{scheme}://{host_path}"
    else:
        if os.name == "nt":
            # Windows can also live with forward slashes
            abs_path = abs_path.replace("\\", "/")
        return _remove_path_dot_segments(abs_path)


#
# Following code is stolen from urllib3.url
#
def _remove_path_dot_segments(path: str) -> str:
    # See http://tools.ietf.org/html/rfc3986#section-5.2.4 for pseudo-code
    segments = path.split("/")  # Turn the path into a list of segments
    output = []  # Initialize the variable to use to store output

    for segment in segments:
        # '.' is the current directory, so ignore it, it is superfluous
        if segment == ".":
            continue
        # Anything other than '..', should be appended to the output
        elif segment != "..":
            output.append(segment)
        # In this case segment == '..', if we can, we should pop the last
        # element
        elif output:
            output.pop()

    # If the path starts with '/' and the output is empty or the first string
    # is non-empty
    if path.startswith("/") and (not output or output[0]):
        output.insert(0, "")

    # If the path starts with '/.' or '/..' ensure we add one more empty
    # string to add a trailing '/'
    if path.endswith(("/.", "/..")):
        output.append("")

    return "/".join(output)


def is_absolute_path(path: str) -> bool:
    """Test whether *path* is an absolute filesystem path or URL."""
    # This is a rather weak test, may be enhanced if desired
    return "//" in path or ":" in path or path.startswith("/")


def get_url_prefix(config: Mapping[str, Any]) -> str:
    """Get the sanitized URL prefix so, if given, it starts with
    a leading slash and ends without one.

    Args:
        config: Server configuration.

    Returns:
        Sanitized URL prefix, may be an empty string.
    """
    return _sanitize_url_prefix(config.get("url_prefix"))


def get_reverse_url_prefix(config: Mapping[str, Any]) -> str:
    """Get the sanitized reverse URL prefix so, if given, it starts with
    a leading slash and ends without one.

    Args:
        config: Server configuration.

    Returns:
        Sanitized URL prefix, may be an empty string.
    """
    return _sanitize_url_prefix(
        config.get("reverse_url_prefix", config.get("url_prefix"))
    )


def _sanitize_url_prefix(url_prefix: Optional[str]) -> str:
    """Get a sanitized URL prefix so, if given, it starts with
    a leading slash and ends without one.

    Args:
        url_prefix: URL prefix path.

    Returns:
        Sanitized URL prefix path, may be an empty string.
    """
    if not url_prefix:
        return ""

    while url_prefix.startswith("//"):
        url_prefix = url_prefix[1:]
    while url_prefix.endswith("/"):
        url_prefix = url_prefix[:-1]

    if url_prefix == "":
        return ""

    if (
        url_prefix.startswith("/")
        or url_prefix.startswith("http://")
        or url_prefix.startswith("https://")
    ):
        return url_prefix

    return "/" + url_prefix



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import warnings
from collections.abc import Collection, Mapping
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import cftime
import dask.array as da
import numpy as np
import pandas as pd
import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.util.assertions import assert_given
from xcube.util.timeindex import ensure_time_index_compatible

Bbox = tuple[float, float, float, float]
TimeRange = Union[
    tuple[Optional[str], Optional[str]],
    tuple[Optional[pd.Timestamp], Optional[pd.Timestamp]],
]


def select_subset(
    dataset: xr.Dataset,
    *,
    var_names: Optional[Collection[str]] = None,
    bbox: Optional[Bbox] = None,
    time_range: Optional[TimeRange] = None,
    grid_mapping: Optional[GridMapping] = None,
):
    """Create a subset from *dataset* given *var_names*,
    *bbox*, *time_range*.

    This is a high-level convenience function that may invoke

    * :func:`select_variables_subset`
    * :func:`select_spatial_subset`
    * :func:`select_temporal_subset`

    Args:
        dataset: The dataset.
        var_names: Optional variable names.
        bbox: Optional bounding box in the dataset's CRS coordinate
            units.
        time_range: Optional time range
        grid_mapping: Optional dataset grid mapping.

    Returns:
        a subset of *dataset*, or unchanged *dataset* if no keyword-
        arguments are used.
    """
    if var_names is not None:
        dataset = select_variables_subset(dataset, var_names=var_names)
    if bbox is not None:
        dataset = select_spatial_subset(
            dataset, xy_bbox=bbox, grid_mapping=grid_mapping
        )
    if time_range is not None:
        dataset = select_temporal_subset(dataset, time_range=time_range)
    return dataset


def select_variables_subset(
    dataset: xr.Dataset, var_names: Optional[Collection[str]] = None
) -> xr.Dataset:
    """Select data variable from given *dataset* and create new dataset.

    Args:
        dataset: The dataset from which to select variables.
        var_names: The names of data variables to select.

    Returns:
        A new dataset. It is empty, if *var_names* is empty. It is
        *dataset*, if *var_names* is None.
    """
    if var_names is None:
        return dataset
    dropped_variables = set(dataset.data_vars.keys()).difference(var_names)
    if not dropped_variables:
        return dataset
    return dataset.drop_vars(dropped_variables)


def select_spatial_subset(
    dataset: xr.Dataset,
    ij_bbox: Optional[tuple[int, int, int, int]] = None,
    ij_border: int = 0,
    xy_bbox: Optional[tuple[float, float, float, float]] = None,
    xy_border: float = 0.0,
    grid_mapping: Optional[GridMapping] = None,
) -> Optional[xr.Dataset]:
    """Select a spatial subset of *dataset* for the
    bounding box *ij_bbox* or *xy_bbox*.

    *ij_bbox* or *xy_bbox* must not be given both.

    Args:
        xy_bbox: The bounding box in x,y coordinates.
        xy_border: Border in units of the x,y coordinates.
        dataset: Source dataset.
        ij_bbox: Bounding box (i_min, i_min, j_max, j_max) in pixel
            coordinates.
        ij_border: Extra border added to *ij_bbox* in number of pixels
        grid_mapping: Optional dataset grid mapping.

    Returns:
        Spatial dataset subset
    """

    if ij_bbox is None and xy_bbox is None:
        raise ValueError("One of ij_bbox and xy_bbox must be given")
    if ij_bbox and xy_bbox:
        raise ValueError("Only one of ij_bbox and xy_bbox can be given")

    if grid_mapping is None:
        grid_mapping = GridMapping.from_dataset(dataset)
    x_name, y_name = grid_mapping.xy_var_names
    x = dataset[x_name]
    y = dataset[y_name]

    if x.ndim == 1 and y.ndim == 1:
        # Hotfix für #981 and #985
        if xy_bbox:
            if y.values[0] < y.values[-1]:
                ds = dataset.sel(
                    **{
                        x_name: slice(xy_bbox[0] - xy_border, xy_bbox[2] + xy_border),
                        y_name: slice(xy_bbox[1] - xy_border, xy_bbox[3] + xy_border),
                    }
                )
            else:
                ds = dataset.sel(
                    **{
                        x_name: slice(xy_bbox[0] - xy_border, xy_bbox[2] + xy_border),
                        y_name: slice(xy_bbox[3] + xy_border, xy_bbox[1] - xy_border),
                    }
                )
            return ds
        else:
            return dataset.isel(
                **{
                    x_name: slice(ij_bbox[0] - ij_border, ij_bbox[2] + ij_border),
                    y_name: slice(ij_bbox[1] - ij_border, ij_bbox[3] + ij_border),
                }
            )
    else:
        if xy_bbox:
            ij_bbox = grid_mapping.ij_bbox_from_xy_bbox(
                xy_bbox, ij_border=ij_border, xy_border=xy_border
            )
            if ij_bbox[0] == -1:
                return None
        width, height = grid_mapping.size
        i_min, j_min, i_max, j_max = ij_bbox
        if i_min > 0 or j_min > 0 or i_max < width - 1 or j_max < height - 1:
            x_dim, y_dim = grid_mapping.xy_dim_names
            i_slice = slice(i_min, i_max + 1)
            j_slice = slice(j_min, j_max + 1)
            return dataset.isel({x_dim: i_slice, y_dim: j_slice})
        return dataset


def select_temporal_subset(
    dataset: xr.Dataset, time_range: TimeRange, time_name: str = "time"
) -> xr.Dataset:
    """Select a temporal subset from *dataset* given *time_range*.

    Args:
        dataset: The dataset. Must include time
        time_range: Time range given as two time stamps (start, end)
            that may be (ISO) strings or datetime objects.
        time_name: optional name of the time coordinate variable.
            Defaults to "time".

    Returns:

    """
    assert_given(time_range, "time_range")
    time_name = time_name or "time"
    if time_name not in dataset:
        raise ValueError(
            f"cannot compute temporal subset: variable"
            f' "{time_name}" not found in dataset'
        )
    time_1, time_2 = time_range
    time_1 = pd.to_datetime(time_1) if time_1 is not None else None
    time_2 = pd.to_datetime(time_2) if time_2 is not None else None
    if time_1 is None and time_2 is None:
        return dataset
    if time_2 is not None:
        delta = time_2 - time_2.floor("1D")
        if delta == pd.Timedelta("0 days 00:00:00"):
            time_2 += pd.Timedelta("1D")
    try:
        time_slice = ensure_time_index_compatible(
            dataset, slice(time_1, time_2), time_name
        )
        return dataset.sel({time_name or "time": time_slice})
    except TypeError:
        calendar = dataset.time.encoding.get("calendar")
        time_1 = cftime.datetime(
            time_1.year, time_1.month, time_1.day, calendar=calendar
        )
        time_2 = cftime.datetime(
            time_2.year, time_2.month, time_2.day, calendar=calendar
        )
        time_slice = ensure_time_index_compatible(
            dataset, slice(time_1, time_2), time_name
        )
        return dataset.sel({time_name or "time": time_slice})


_PREDICATE_SIGNATURE = "predicate(slice_array: xr.DataArray, slice_info: Dict) -> bool"

Predicate = Callable[[xr.DataArray, dict[str, Any]], bool]


def select_label_subset(
    dataset: xr.Dataset,
    dim: str,
    predicate: Union[Predicate, Mapping[str, Predicate]],
    use_dask: bool = False,
):
    """Select the labels in *dataset* along a given dimension *dim*
    using a predicate function *predicate* that is called for
    all variable slices for a current label.

    The *predicate* can also be provided as a mapping
    from variable names to dedicated predicate functions.

    The predicate function is called for all *dim* labels in *dataset*
    and for every variable that contains *dim*.

    If *predicate* returns False for any given label,
    that label will be dropped from dimension *dim*.

    Predicate functions are defined as follows:::

        def predicate(slice_array: xr.DataArray, slice_info: Dict) -> bool:
            ...

    Here, *slice_array* is a variable's array slice for the given label.
    The argument *slice_info* is a dictionary that contains the
    following keys:

    * var: str - name of the current variable.
    * dim: str - value of *dim*.
    * index: int - value for the current index within dimension *dim*.
    * label: Optional[xr.DataArray] - value for the current label
      within dimension *dim*.

    Note, the value of "label" will be None, if *dataset*
    does not contain a 1D-coordinate variable named *dim*.

    The following example selects only time labels
    from a 3-D (time, y, x) cube where the 2-D (y, x) images
    of variable "CHL" comprises more than 50% valid values:::

        >>> chl_data = np.random.random((5, 10, 20))
        >>> chl_data = np.where(chl_data > 0.5, chl_data, np.nan)
        >>> ds = xr.Dataset({"CHL": (["time", "y", "x"], chl_data)})
        >>>
        >>> def is_valid_slice(slice_array, slice_label):
        >>>     return np.sum(np.isnan(slice_array)) / slice_array.size <= 0.5
        >>>
        >>> ds_subset = select_label_subset(ds, "time",
        >>>                                 predicate={"CHL": is_valid_slice})

    Args:
        dataset: The dataset.
        dim: The name of the dimension from which to select the labels.
        predicate: The predicate function or a mapping from variable
            names to variable-specific predicate functions.
        use_dask: Whether to use a Dask graph that will compute the
            validity of labels in parallel. For a large number of
            labels, very complex Dask graphs will result (every label is
            a node) whose overhead may compensate the performance gain.

    Returns:
        A new dataset with labels along *dim* selected by the
        *predicate*. If all labels are selected, *dataset* is returned
        without change.
    """
    if callable(predicate):
        predicate_lookup = {
            var_name: predicate
            for var_name, var in dataset.data_vars.items()
            if dim in var.dims
        }
    elif isinstance(predicate, Mapping):
        predicate_lookup = predicate
        for var_name, var_predicate in predicate_lookup.items():
            if not callable(var_predicate):
                raise TypeError(
                    f"predicate for variable {var_name!r}"
                    f" must be callable with"
                    f" signature {_PREDICATE_SIGNATURE}"
                )
    else:
        raise TypeError(
            f"predicate must be callable with signature {_PREDICATE_SIGNATURE}"
        )

    num_labels = dataset.sizes[dim]

    valid_mask = [
        _is_label_valid(dataset, predicate_lookup, dim, index)
        for index in range(num_labels)
    ]

    if use_dask:
        valid_mask = da.stack(valid_mask).compute()

    dropped_indexes = [i for i in range(num_labels) if not valid_mask[i]]
    if not dropped_indexes:
        return dataset

    return dataset.drop_isel({dim: dropped_indexes})


def _is_label_valid(
    dataset: xr.Dataset, predicate_lookup: Mapping[str, Predicate], dim: str, index: int
) -> da.Array:
    label = dataset[dim][index] if dim in dataset else None
    results: list[da.Array] = []
    for var_name, var in dataset.data_vars.items():
        if dim in var.dims:
            predicate = predicate_lookup.get(var_name)
            if predicate is not None:
                slice_array = var.isel({dim: index})
                slice_info = dict(var=var_name, dim=dim, index=index, label=label)
                result = predicate(slice_array, slice_info)
                if isinstance(result, xr.DataArray):
                    result = result.data
                if isinstance(result, da.Array):
                    results.append(result)
                else:
                    results.append(da.from_array(result))
    if len(results) == 0:
        return da.from_array(True)
    elif len(results) == 1:
        return results[0]
    else:
        return da.all(da.stack(results))



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import xarray as xr

from xcube.core.verify import assert_cube


def vars_to_dim(
    cube: xr.Dataset,
    dim_name: str = "var",
    var_name="data",
    cube_asserted: bool = False,
):
    """Convert data variables into a dimension.

    Args:
        cube: The xcube dataset.
        dim_name: The name of the new dimension and coordinate variable.
            Defaults to 'var'.
        var_name: The name of the new, single data variable. Defaults to
            'data'.
        cube_asserted: If False, *cube* will be verified, otherwise it
            is expected to be a valid cube.

    Returns:
        A new xcube dataset with data variables turned into a new
        dimension.
    """

    if not cube_asserted:
        assert_cube(cube)

    if var_name == dim_name:
        raise ValueError("var_name must be different from dim_name")

    data_var_names = [data_var_name for data_var_name in cube.data_vars]
    if not data_var_names:
        raise ValueError("cube must not be empty")

    da = xr.concat([cube[data_var_name] for data_var_name in data_var_names], dim_name)
    new_coord_var = xr.DataArray(data_var_names, dims=[dim_name])
    da = da.assign_coords(**{dim_name: new_coord_var})

    return xr.Dataset(dict(**{var_name: da}))



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import fnmatch
from typing import Optional, Set

from xcube.server.api import ApiError

READ_ALL_DATASETS_SCOPE = "read:dataset:*"
READ_ALL_VARIABLES_SCOPE = "read:variable:*"


def assert_scopes(
    required_scopes: set[str],
    granted_scopes: Optional[set[str]],
    is_substitute: bool = False,
):
    """Assert scopes.
    Raise ServiceAuthError if one of *required_scopes* is
    not in *granted_scopes*.

    Args:
        required_scopes: The list of required scopes
        granted_scopes: The set of granted scopes. If user is not
            authenticated, its value is None.
        is_substitute: True, if the resource to be checked is a
            substitute.
    """
    missing_scope = _get_missing_scope(
        required_scopes, granted_scopes, is_substitute=is_substitute
    )
    if missing_scope is not None:
        raise ApiError.Unauthorized(f'Missing permission "{missing_scope}"')


def check_scopes(
    required_scopes: set[str],
    granted_scopes: Optional[set[str]],
    is_substitute: bool = False,
) -> bool:
    """Check scopes.

    This function is used to filter out a resource's sub-resources for
    which a given client has no permission.

    If one of *required_scopes* is not in *granted_scopes*, fail.
    If *granted_scopes* exists and *is_substitute*, fail too.
    Else succeed.

    Args:
        required_scopes: The list of required scopes
        granted_scopes: The set of granted scopes. If user is not
            authenticated, its value is None.
        is_substitute: True, if the resource to be checked is a
            substitute.

    Returns:
        True, if scopes are ok.
    """
    return (
        _get_missing_scope(required_scopes, granted_scopes, is_substitute=is_substitute)
        is None
    )


def _get_missing_scope(
    required_scopes: set[str],
    granted_scopes: Optional[set[str]],
    is_substitute: bool = False,
) -> Optional[str]:
    """Return the first required scope that is
    fulfilled by any granted scope

    Args:
        required_scopes: The list of required scopes
        granted_scopes: The set of granted scopes. If user is not
            authenticated, its value is None.
        is_substitute: True, if the resource to be checked is a
            substitute.

    Returns:
        The missing scope.
    """
    is_authenticated = granted_scopes is not None
    if is_authenticated:
        for required_scope in required_scopes:
            required_permission_given = False
            for granted_scope in granted_scopes:
                if required_scope == granted_scope or fnmatch.fnmatch(
                    required_scope, granted_scope
                ):
                    # If any granted scope matches, we can stop
                    required_permission_given = True
                    break
            if not required_permission_given:
                # The required scope is not a granted scope --> fail
                return required_scope

        # If we end here, required_scopes are either empty or satisfied
        if is_substitute:
            # All required scopes are satisfied, now fail for
            # substitute resources (e.g. demo resources) as there
            # is usually a better (non-demo) resource that replaces it.
            # Return missing scope (dummy, not used) --> fail
            return READ_ALL_DATASETS_SCOPE

    elif required_scopes:
        # We require scopes but have no granted scopes
        if not is_substitute:
            # ...and resource is not a substitute --> fail
            return next(iter(required_scopes))

    # All ok.
    return None



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

# noinspection PyUnresolvedReferences
# noinspection PyUnresolvedReferences
from xcube.core.byoa import CodeConfig, FileSet

# noinspection PyUnresolvedReferences
from xcube.core.store.descriptor import DatasetDescriptor

from .config import CallbackConfig, CubeConfig, InputConfig, OutputConfig
from .error import CubeGeneratorError
from .generator import CubeGenerator
from .local.generator import LocalCubeGenerator
from .processor import (
    METHOD_NAME_DATASET_PROCESSOR,
    METHOD_NAME_PARAMS_SCHEMA_GETTER,
    DatasetProcessor,
)
from .remote.config import ServiceConfig, ServiceConfigLike
from .remote.generator import RemoteCubeGenerator
from .remote.response import CostEstimation, CubeInfoWithCosts, CubeInfoWithCostsResult
from .request import CubeGeneratorRequest, CubeGeneratorRequestLike
from .response import CubeGeneratorResult, CubeInfo, CubeInfoResult, CubeReference



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Any, Callable, List, Optional, Union

from xcube.util.jsonschema import JsonObjectSchema, JsonSchema

from .info import PyType
from .registry import OP_REGISTRY, OpRegistry


def operation(
    _op: Optional[Callable] = None,
    params_schema: Optional[JsonObjectSchema] = None,
    op_registry: OpRegistry = OP_REGISTRY,
):
    """Decorator that registers a function as an operation.

    Args:
        _op: the function to register as an operation
        params_schema: JSON Schema for the function's parameters (a JSON
            object schema mapping parameter names to their individual
            schemas)
        op_registry: the registry in which to register the operation

    Returns:
        the decorated operation, if an operation was supplied;
        otherwise, a decorator function
    """

    def decorator(op: Callable):
        _assert_decorator_target_ok("operation", op)
        op_info = op_registry.register_op(op)
        if params_schema is not None:
            op_info.update_params_schema(params_schema.to_dict())
        return op

    if _op is None:
        return decorator
    else:
        return decorator(_op)


def op_param(
    name: str,
    json_type: Optional[Union[str, list[str]]] = None,
    py_type: Optional[PyType] = None,
    title: Optional[str] = None,
    description: Optional[str] = None,
    default: Optional[Any] = None,
    required: Optional[bool] = None,
    schema: Optional[JsonSchema] = None,
    op_registry: OpRegistry = OP_REGISTRY,
):
    """Decorator that adds schema information to the operation parameter given
    by *name*.

    See also
    https://json-schema.org/draft/2020-12/json-schema-validation.html#name-a-vocabulary-for-basic-meta

    Args:
        name: name of the parameter to apply schema information to
        json_type: JSON Schema type of the parameter
        py_type: Python type of the parameter
        title: title of the parameter
        description: description of the parameter
        default: default value for the parameter
        required: whether the parameter is required
        schema: JSON Schema describing the parameter
        op_registry: registry in which to register the operation

    Returns:
        parameterized decorator for a compute operation function
    """

    def decorator(op: Callable):
        _assert_decorator_target_ok("op_param", op)
        op_info = op_registry.register_op(op)
        op_param_schema = {}
        if schema is not None:
            op_param_schema.update(schema.to_dict())
        if py_type is not None:
            op_info.set_param_py_type(name, py_type)
        if json_type is not None:
            op_param_schema.update({"type": json_type})
        if title is not None:
            op_param_schema.update({"title": title})
        if description is not None:
            op_param_schema.update({"description": description})
        if default is not None:
            op_param_schema.update({"default": default})
        if op_param_schema:
            op_info.update_param_schema(name, op_param_schema)
        if required is not None:
            required_set = set(op_info.params_schema.get("required", []))
            if required and name not in required_set:
                required_set.add(name)
            elif not required and name in required_set:
                required_set.remove(name)
            op_info.update_params_schema({"required": list(required_set)})
        return op

    return decorator


def _assert_decorator_target_ok(decorator_name: str, target: Any):
    if not callable(target):
        raise TypeError(f"decorator {decorator_name!r} can be used with callables only")



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .routes import (
    PATH_PARAM_X,
    PATH_PARAM_Y,
    PATH_PARAM_Z,
    QUERY_PARAM_FORMAT,
    QUERY_PARAM_RETINA,
    QUERY_PARAM_TIME,
    api,
)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from abc import abstractmethod

import geopandas as gpd
import pandas as pd

from xcube.util.assertions import assert_instance
from xcube.util.fspath import is_local_fs
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.temp import new_temp_file

from ...datatype import GEO_DATA_FRAME_TYPE, DataType
from ..accessor import FsDataAccessor


class GeoDataFrameFsDataAccessor(FsDataAccessor):
    """Extension name: 'geodataframe:<format_id>:<protocol>'."""

    @classmethod
    def get_data_type(cls) -> DataType:
        return GEO_DATA_FRAME_TYPE

    @classmethod
    @abstractmethod
    def get_driver_name(cls) -> str:
        """Get the GeoDataFrame I/O driver name"""

    def get_open_data_params_schema(self, data_id: str = None) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                storage_options=self.get_storage_options_schema(),
                # TODO: add more, see https://geopandas.org/io.html
            ),
        )

    def open_data(self, data_id: str, **open_params) -> gpd.GeoDataFrame:
        # TODO: implement me correctly,
        #  this is not valid for shapefile AND geojson
        fs, root, open_params = self.load_fs(open_params)
        is_local = is_local_fs(fs)
        if is_local:
            file_path = data_id
        else:
            _, file_path = new_temp_file()
            fs.get_file(data_id, file_path)
        return gpd.read_file(file_path, driver=self.get_driver_name(), **open_params)

    def get_write_data_params_schema(self) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                storage_options=self.get_storage_options_schema(),
                # TODO: add more, see https://geopandas.org/io.html
            ),
        )

    def write_data(self, data: gpd.GeoDataFrame, data_id: str, **write_params) -> str:
        # TODO: implement me correctly,
        #  this is not valid for shapefile AND geojson
        assert_instance(data, (gpd.GeoDataFrame, pd.DataFrame), "data")
        fs, root, write_params = self.load_fs(write_params)
        is_local = is_local_fs(fs)
        if is_local:
            file_path = data_id
        else:
            _, file_path = new_temp_file()
        data.to_file(file_path, driver=self.get_driver_name(), **write_params)
        if not is_local:
            fs.put_file(file_path, data_id)
        return data_id


class GeoDataFrameShapefileFsDataAccessor(GeoDataFrameFsDataAccessor):
    """Extension name: 'geodataframe:shapefile:<protocol>'."""

    @classmethod
    def get_format_id(cls) -> str:
        return "shapefile"

    @classmethod
    def get_driver_name(cls) -> str:
        return "ESRI Shapefile"


class GeoDataFrameGeoJsonFsDataAccessor(GeoDataFrameFsDataAccessor):
    """Extension name: 'geodataframe:geojson:<protocol>'."""

    @classmethod
    def get_format_id(cls) -> str:
        return "geojson"

    @classmethod
    def get_driver_name(cls) -> str:
        return "GeoJSON"



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import threading
import uuid
from abc import ABCMeta, abstractmethod
from collections.abc import Mapping
from typing import Any, Dict, Optional

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.util.assertions import assert_instance

from .abc import MultiLevelDataset


class LazyMultiLevelDataset(MultiLevelDataset, metaclass=ABCMeta):
    """A multi-level dataset where each level dataset is lazily retrieved,
    i.e. read or computed by the abstract method
    ``get_dataset_lazily(index, **kwargs)``.

    Args:
        ds_id: Optional dataset identifier.
        parameters: Optional keyword arguments that will be passed to
            the ``get_dataset_lazily`` method.
    """

    def __init__(
        self,
        grid_mapping: Optional[GridMapping] = None,
        num_levels: Optional[int] = None,
        ds_id: Optional[str] = None,
        parameters: Optional[Mapping[str, Any]] = None,
    ):
        if grid_mapping is not None:
            assert_instance(grid_mapping, GridMapping, name="grid_mapping")
        if ds_id is not None:
            assert_instance(ds_id, str, name="ds_id")
        self._grid_mapping = grid_mapping
        self._num_levels = num_levels
        self._ds_id = ds_id
        self._level_datasets: dict[int, xr.Dataset] = {}
        self._parameters = parameters or {}
        self._lock = threading.RLock()

    @property
    def ds_id(self) -> str:
        if self._ds_id is None:
            with self._lock:
                self._ds_id = str(uuid.uuid4())
        return self._ds_id

    @ds_id.setter
    def ds_id(self, ds_id: str):
        assert_instance(ds_id, str, name="ds_id")
        self._ds_id = ds_id

    @property
    def grid_mapping(self) -> GridMapping:
        if self._grid_mapping is None:
            with self._lock:
                self._grid_mapping = self._get_grid_mapping_lazily()
        return self._grid_mapping

    @property
    def num_levels(self) -> int:
        if self._num_levels is None:
            with self._lock:
                self._num_levels = self._get_num_levels_lazily()
        return self._num_levels

    @property
    def lock(self) -> threading.RLock:
        """Get the reentrant lock used by this object to synchronize
        lazy instantiation of properties.
        """
        return self._lock

    def get_dataset(self, index: int) -> xr.Dataset:
        """Get or compute the dataset for the level at given *index*.

        Args:
            index: the level index

        Returns:
            the dataset for the level at *index*.
        """
        if index not in self._level_datasets:
            with self._lock:
                # noinspection PyTypeChecker
                level_dataset = self._get_dataset_lazily(index, self._parameters)
                self.set_dataset(index, level_dataset)
        # noinspection PyTypeChecker
        return self._level_datasets[index]

    def set_dataset(self, index: int, level_dataset: xr.Dataset):
        """Set the dataset for the level at given *index*.

        Callers need to ensure that the given *level_dataset*
        has the correct spatial dimension sizes for the
        given level at *index*.

        Args:
            index: the level index
            level_dataset: the dataset for the level at *index*.
        """
        with self._lock:
            self._level_datasets[index] = level_dataset

    @abstractmethod
    def _get_num_levels_lazily(self) -> int:
        """Retrieve, i.e. read or compute, the number of levels.

        Returns:
            the number of dataset levels.
        """

    @abstractmethod
    def _get_dataset_lazily(self, index: int, parameters: dict[str, Any]) -> xr.Dataset:
        """Retrieve, i.e. read or compute, the dataset for the
        level at given *index*.

        Args:
            index: the level index
            parameters: *parameters* keyword argument that was passed to
                constructor.

        Returns:
            the dataset for the level at *index*.
        """

    def _get_grid_mapping_lazily(self) -> GridMapping:
        """Retrieve, i.e. read or compute, the tile grid used
        by the multi-level dataset.

        Returns:
            the dataset for the level at *index*.
        """
        return GridMapping.from_dataset(self.get_dataset(0))

    def close(self):
        with self._lock:
            for dataset in self._level_datasets.values():
                if dataset is not None:
                    dataset.close()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import os
import shutil
import warnings
from abc import ABCMeta, abstractmethod
from collections.abc import Iterable, Mapping
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import botocore.exceptions
import pandas as pd
import s3fs
import urllib3.util
import xarray as xr
import zarr
from deprecated import deprecated

from xcube.constants import (
    EXTENSION_POINT_DATASET_IOS,
    FORMAT_NAME_CSV,
    FORMAT_NAME_MEM,
    FORMAT_NAME_NETCDF4,
    FORMAT_NAME_ZARR,
)
from xcube.core.timeslice import (
    append_time_slice,
    insert_time_slice,
    replace_time_slice,
)
from xcube.core.verify import assert_cube
from xcube.util.plugin import ExtensionComponent, get_extension_registry

_DEPRECATION_REASON = "Functionality is redundant. Use xcube.core.store API instead."
_DEPRECATION_VERSION = "0.12.1"


# Note, we cannot remove this deprecated code as long as
# xcube.core.xarray.DatasetAccessor.open() is using it.
@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def open_cube(input_path: str, format_name: str = None, **kwargs) -> xr.Dataset:
    """Open a xcube dataset from *input_path*.
    If *format* is not provided it will be guessed from *input_path*.

    Args:
        input_path: input path
        format_name: format, e.g. "zarr" or "netcdf4"
        **kwargs: format-specific keyword arguments

    Returns:
        xcube dataset
    """
    return open_dataset(input_path, format_name=format_name, is_cube=True, **kwargs)


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def write_cube(
    cube: xr.Dataset,
    output_path: str,
    format_name: str = None,
    cube_asserted: bool = False,
    **kwargs,
) -> xr.Dataset:
    """Write a xcube dataset to *output_path*.
    If *format* is not provided it will be guessed from *output_path*.

    Args:
        cube: xcube dataset to be written.
        output_path: output path
        format_name: format, e.g. "zarr" or "netcdf4"
        **kwargs: format-specific keyword arguments
        cube_asserted: If False, *cube* will be verified, otherwise it
            is expected to be a valid cube.

    Returns:
        xcube dataset *cube*
    """
    if not cube_asserted:
        assert_cube(cube)
    return write_dataset(cube, output_path, format_name=format_name, **kwargs)


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def open_dataset(
    input_path: str, format_name: str = None, is_cube: bool = False, **kwargs
) -> xr.Dataset:
    """Open a dataset from *input_path*.
    If *format* is not provided it will be guessed from *output_path*.

    Args:
        input_path: input path
        format_name: format, e.g. "zarr" or "netcdf4"
        is_cube: Whether a ValueError will be raised, if the dataset
            read from *input_path* is not a xcube dataset.
        **kwargs: format-specific keyword arguments

    Returns:
        dataset object
    """
    format_name = format_name if format_name else guess_dataset_format(input_path)
    if format_name is None:
        raise ValueError("Unknown input format")
    dataset_io = find_dataset_io(format_name, modes=["r"])
    if dataset_io is None:
        raise ValueError(f"Unknown input format {format_name!r} for {input_path}")
    dataset = dataset_io.read(input_path, **kwargs)
    if is_cube:
        assert_cube(dataset)
    return dataset


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def write_dataset(
    dataset: xr.Dataset, output_path: str, format_name: str = None, **kwargs
) -> xr.Dataset:
    """Write dataset to *output_path*.
    If *format* is not provided it will be guessed from *output_path*.

    Args:
        dataset: Dataset to be written.
        output_path: output path
        format_name: format, e.g. "zarr" or "netcdf4"
        **kwargs: format-specific keyword arguments

    Returns:
        the input dataset
    """
    format_name = format_name if format_name else guess_dataset_format(output_path)
    if format_name is None:
        raise ValueError("Unknown output format")
    dataset_io = find_dataset_io(format_name, modes=["w"])
    if dataset_io is None:
        raise ValueError(f"Unknown output format {format_name!r} for {output_path}")
    dataset_io.write(dataset, output_path, **kwargs)
    return dataset


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
class DatasetIO(ExtensionComponent, metaclass=ABCMeta):
    """An abstract base class that represents dataset input/output.

    Args:
        name: A unique dataset I/O identifier.
    """

    def __init__(self, name: str):
        super().__init__(EXTENSION_POINT_DATASET_IOS, name)

    @property
    def description(self) -> str:
        """Returns:
        A description for this input processor
        """
        return self.get_metadata_attr("description", "")

    @property
    def ext(self) -> str:
        """The primary filename extension used by this dataset I/O."""
        return self.get_metadata_attr("ext", "")

    @property
    def modes(self) -> set[str]:
        """A set describing the modes of this dataset I/O.
        Must be one or more of "r" (read), "w" (write), and "a" (append).
        """
        return self.get_metadata_attr("modes", set())

    @abstractmethod
    def fitness(self, path: str, path_type: str = None) -> float:
        """Compute a fitness of this dataset I/O in the interval [0 to 1]
        for reading/writing from/to the given *path*.

        Args:
            path: The path or URL.
            path_type: Either "file", "dir", "url", or None.

        Returns:
            the chance in range [0 to 1]
        """
        return 0.0

    def read(self, input_path: str, **kwargs) -> xr.Dataset:
        """Read a dataset from *input_path* using format-specific read parameters *kwargs*."""
        raise NotImplementedError()

    def write(self, dataset: xr.Dataset, output_path: str, **kwargs):
        """ "Write *dataset* to *output_path* using format-specific write parameters *kwargs*."""
        raise NotImplementedError()

    def append(self, dataset: xr.Dataset, output_path: str, **kwargs):
        """ "Append *dataset* to existing *output_path* using format-specific write parameters *kwargs*."""
        raise NotImplementedError()

    def insert(self, dataset: xr.Dataset, index: int, output_path: str, **kwargs):
        """ "Insert *dataset* at *index* into existing *output_path* using format-specific write parameters *kwargs*."""
        raise NotImplementedError()

    def replace(self, dataset: xr.Dataset, index: int, output_path: str, **kwargs):
        """ "Replace *dataset* at *index* in existing *output_path* using format-specific write parameters *kwargs*."""
        raise NotImplementedError()

    def update(self, output_path: str, global_attrs: dict[str, Any] = None, **kwargs):
        """ "Update *dataset* at *output_path* using format-specific open parameters *kwargs*."""
        raise NotImplementedError()


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def get_extension(name: str):
    return get_extension_registry().get_extension(EXTENSION_POINT_DATASET_IOS, name)


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def find_dataset_io_by_name(name: str):
    extension = get_extension(name)
    if not extension:
        return None
    return extension.component


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def find_dataset_io(
    format_name: str, modes: Iterable[str] = None, default: DatasetIO = None
) -> Optional[DatasetIO]:
    modes = set(modes) if modes else None
    format_name = format_name.lower()
    dataset_ios = get_extension_registry().find_components(EXTENSION_POINT_DATASET_IOS)
    for dataset_io in dataset_ios:
        # noinspection PyUnresolvedReferences
        if format_name == dataset_io.name.lower():
            # noinspection PyTypeChecker
            if not modes or modes.issubset(dataset_io.modes):
                return dataset_io
    for dataset_io in dataset_ios:
        # noinspection PyUnresolvedReferences
        if format_name == dataset_io.ext.lower():
            # noinspection PyTypeChecker
            if not modes or modes.issubset(dataset_io.modes):
                return dataset_io
    return default


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def guess_dataset_format(path: str) -> Optional[str]:
    """Guess a dataset format for a file system path or URL given by *path*.

    Args:
        path: A file system path or URL.

    Returns:
        The name of a dataset format guessed from *path*.
    """
    dataset_io_fitness_list = guess_dataset_ios(path)
    if dataset_io_fitness_list:
        return dataset_io_fitness_list[0][0].name
    return None


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def guess_dataset_ios(path: str) -> list[tuple[DatasetIO, float]]:
    """Guess suitable DatasetIO objects for a file system path or URL given by *path*.
    Returns a list of (DatasetIO, fitness) tuples, sorted by descending fitness values.
    Fitness values are in the interval (0, 1].
    The first entry is the most appropriate DatasetIO object.

    Args:
        path: A file system path or URL.

    Returns:
        A list of (DatasetIO, fitness) tuples.
    """
    if os.path.isfile(path):
        input_type = "file"
    elif os.path.isdir(path):
        input_type = "dir"
    elif path.find("://") > 0:
        input_type = "url"
    else:
        input_type = None

    dataset_ios = get_extension_registry().find_components(EXTENSION_POINT_DATASET_IOS)

    dataset_io_fitness_list = []
    for dataset_io in dataset_ios:
        fitness = dataset_io.fitness(path, path_type=input_type)
        if fitness > 0.0:
            dataset_io_fitness_list.append((dataset_io, fitness))

    dataset_io_fitness_list.sort(key=lambda item: -item[1])
    return dataset_io_fitness_list


def _get_ext(path: str) -> Optional[str]:
    _, ext = os.path.splitext(path)
    return ext.lower()


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def query_dataset_io(filter_fn: Callable[[DatasetIO], bool] = None) -> list[DatasetIO]:
    dataset_ios = get_extension_registry().find_components(EXTENSION_POINT_DATASET_IOS)
    if filter_fn is None:
        return dataset_ios
    return list(filter(filter_fn, dataset_ios))


# noinspection PyAbstractClass
@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
class MemDatasetIO(DatasetIO):
    """An in-memory  dataset I/O. Keeps all datasets in a dictionary.

    Args:
        datasets: The initial datasets as a path to dataset mapping.
    """

    def __init__(self, datasets: dict[str, xr.Dataset] = None):
        super().__init__(FORMAT_NAME_MEM)
        self._datasets = datasets or {}

    @property
    def datasets(self) -> dict[str, xr.Dataset]:
        return self._datasets

    def fitness(self, path: str, path_type: str = None) -> float:
        if path in self._datasets:
            return 1.0
        ext_value = _get_ext(path) == ".mem"
        type_value = 0.0
        return (3 * ext_value + type_value) / 4

    def read(self, path: str, **kwargs) -> xr.Dataset:
        if path in self._datasets:
            return self._datasets[path]
        raise FileNotFoundError(path)

    def write(self, dataset: xr.Dataset, path: str, **kwargs):
        self._datasets[path] = dataset

    def append(self, dataset: xr.Dataset, path: str, **kwargs):
        if path in self._datasets:
            old_ds = self._datasets[path]
            # noinspection PyTypeChecker
            self._datasets[path] = xr.concat(
                [old_ds, dataset],
                dim="time",
                data_vars="minimal",
                coords="minimal",
                compat="equals",
            )
        else:
            self._datasets[path] = dataset.copy()

    def update(self, output_path: str, global_attrs: dict[str, Any] = None, **kwargs):
        if global_attrs:
            ds = self._datasets[output_path]
            ds.attrs.update(global_attrs)


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
class Netcdf4DatasetIO(DatasetIO):
    """A dataset I/O that reads from / writes to NetCDF files."""

    def __init__(self):
        super().__init__(FORMAT_NAME_NETCDF4)

    def fitness(self, path: str, path_type: str = None) -> float:
        ext = _get_ext(path)
        ext_value = ext in {".nc", ".hdf", ".h5"}
        type_value = 0.0
        if path_type == "file":
            type_value = 1.0
        elif path_type is None:
            type_value = 0.5
        else:
            ext_value = 0.0
        return (3 * ext_value + type_value) / 4

    def read(self, input_path: str, **kwargs) -> xr.Dataset:
        return xr.open_dataset(input_path, **kwargs)

    def write(self, dataset: xr.Dataset, output_path: str, **kwargs):
        dataset.to_netcdf(output_path)

    def append(self, dataset: xr.Dataset, output_path: str, **kwargs):
        import os

        temp_path = output_path + ".temp.nc"
        os.rename(output_path, temp_path)
        old_ds = xr.open_dataset(temp_path)
        new_ds = xr.concat(
            [old_ds, dataset],
            dim="time",
            data_vars="minimal",
            coords="minimal",
            compat="equals",
        )
        # noinspection PyUnresolvedReferences
        new_ds.to_netcdf(output_path)
        old_ds.close()
        rimraf(temp_path)

    def update(self, output_path: str, global_attrs: dict[str, Any] = None, **kwargs):
        if global_attrs:
            import netCDF4

            ds = netCDF4.Dataset(output_path, "r+")
            ds.setncatts(global_attrs)
            ds.close()


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
class ZarrDatasetIO(DatasetIO):
    """A dataset I/O that reads from / writes to Zarr directories or archives."""

    def __init__(self):
        super().__init__(FORMAT_NAME_ZARR)

    def fitness(self, path: str, path_type: str = None) -> float:
        ext = _get_ext(path)
        ext_value = 0.0
        type_value = 0.0
        if ext == ".zarr":
            ext_value = 1.0
            if path_type == "dir":
                type_value = 1.0
            elif path_type == "url" or path_type is None:
                type_value = 0.5
            else:
                ext_value = 0.0
        else:
            lower_path = path.lower()
            if lower_path.endswith(".zarr.zip"):
                ext_value = 1.0
                if path_type == "file":
                    type_value = 1.0
                elif path_type is None:
                    type_value = 0.5
                else:
                    ext_value = 0.0
            else:
                if path_type == "dir":
                    type_value = 1.0
                elif path_type == "url":
                    type_value = 0.5
        return (3 * ext_value + type_value) / 4

    def read(
        self,
        path: str,
        s3_kwargs: dict[str, Any] = None,
        s3_client_kwargs: dict[str, Any] = None,
        max_cache_size: int = None,
        **kwargs,
    ) -> xr.Dataset:
        """Read dataset from some Zarr storage.

        Args:
            path: File path or object storage URL.
            s3_kwargs: if *path* is an object storage URL, keyword-
                arguments passed to S3 file system, that is
                ``s3fs.S3FileSystem(**s3_kwargs, ...)``.
            s3_client_kwargs: if *path* is an object storage URL,
                keyword-arguments passed to S3 (boto3) client, that is
                ``s3fs.S3FileSystem(...,
                client_kwargs=s3_client_kwargs)``.
            max_cache_size: if this is a positive integer, the store
                will be wrapped in an in-memory cache, that is ``store =
                zarr.LRUStoreCache(store, max_size=max_cache_size)``.
            **kwargs: Keyword-arguments passed to xarray Zarr adapter,
                that is ``xarray.open_zarr(..., **kwargs)``. In
                addition, the parameter **

        Returns:

        """
        path_or_store = path
        consolidated = False
        if isinstance(path, str):
            path_or_store, consolidated = get_path_or_s3_store(
                path_or_store,
                s3_kwargs=s3_kwargs,
                s3_client_kwargs=s3_client_kwargs,
                mode="r",
            )
            if max_cache_size is not None and max_cache_size > 0:
                path_or_store = zarr.LRUStoreCache(
                    path_or_store, max_size=max_cache_size
                )
        return xr.open_zarr(path_or_store, consolidated=consolidated, **kwargs)

    def write(
        self,
        dataset: xr.Dataset,
        output_path: str,
        compressor: dict[str, Any] = None,
        chunksizes: dict[str, int] = None,
        packing: dict[str, dict[str, Any]] = None,
        s3_kwargs: dict[str, Any] = None,
        s3_client_kwargs: dict[str, Any] = None,
        **kwargs,
    ):
        path_or_store, _ = get_path_or_s3_store(
            output_path,
            s3_kwargs=s3_kwargs,
            s3_client_kwargs=s3_client_kwargs,
            mode="w",
        )
        encoding = self._get_write_encodings(dataset, compressor, chunksizes, packing)
        consolidated = kwargs.pop("consolidated", True)
        dataset.to_zarr(
            path_or_store,
            mode="w",
            encoding=encoding,
            consolidated=consolidated,
            **kwargs,
        )

    @classmethod
    def _get_write_encodings(cls, dataset, compressor, chunksizes, packing):
        encoding = None
        if chunksizes:
            encoding = {}
            for var_name in dataset.data_vars:
                var = dataset[var_name]
                chunks: list[int] = []
                for i in range(len(var.dims)):
                    dim_name = var.dims[i]
                    if dim_name in chunksizes:
                        chunks.append(chunksizes[dim_name])
                    else:
                        chunks.append(var.shape[i])
                encoding[var_name] = dict(chunks=chunks)
        if packing:
            if encoding:
                for var_name in packing.keys():
                    if var_name in encoding.keys():
                        encoding[var_name].update(dict(packing[var_name]))
                    else:
                        encoding[var_name] = dict(packing[var_name])
            else:
                encoding = {}
                for var_name in packing.keys():
                    encoding[var_name] = dict(packing[var_name])

        if compressor:
            compressor = zarr.Blosc(**compressor)

            if encoding:
                for var_name in encoding.keys():
                    encoding[var_name].update(compressor=compressor)
            else:
                encoding = {
                    var_name: dict(compressor=compressor)
                    for var_name in dataset.data_vars
                }
        return encoding

    def append(self, dataset: xr.Dataset, output_path: str, **kwargs):
        append_time_slice(output_path, dataset)

    def insert(self, dataset: xr.Dataset, index: int, output_path: str, **kwargs):
        insert_time_slice(output_path, index, dataset)

    def replace(self, dataset: xr.Dataset, index: int, output_path: str, **kwargs):
        replace_time_slice(output_path, index, dataset)

    def update(self, output_path: str, global_attrs: dict[str, Any] = None, **kwargs):
        if global_attrs:
            import zarr

            ds = zarr.open_group(output_path, mode="r+", **kwargs)
            ds.attrs.update(global_attrs)
            zarr.consolidate_metadata(output_path)


# noinspection PyAbstractClass
@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
class CsvDatasetIO(DatasetIO):
    """A dataset I/O that reads from / writes to CSV files."""

    def __init__(self):
        super().__init__(FORMAT_NAME_CSV)

    def fitness(self, path: str, path_type: str = None) -> float:
        if path_type == "dir":
            return 0.0
        ext = _get_ext(path)
        ext_value = {".csv": 1.0, ".txt": 0.5, ".dat": 0.2}.get(ext, 0.1)
        type_value = {"file": 1.0, "url": 0.5, None: 0.5}.get(path_type, 0.0)
        return (3 * ext_value + type_value) / 4

    def read(self, path: str, **kwargs) -> xr.Dataset:
        return xr.Dataset.from_dataframe(pd.read_csv(path, **kwargs))

    def write(self, dataset: xr.Dataset, output_path: str, **kwargs):
        dataset.to_dataframe().to_csv(output_path, **kwargs)


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def rimraf(*paths: str):
    """The UNIX command `rm -rf` for xcube.
    Recursively remove directory or single file.

    Args:
        *paths: one or more directories or files
    """
    for path in paths:
        if os.path.isdir(path):
            try:
                shutil.rmtree(path)
            except OSError:
                warnings.warn(f"failed to remove file {path}")
        elif os.path.isfile(path):
            try:
                os.remove(path)
            except OSError:
                warnings.warn(f"failed to remove file {path}")
                pass


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def get_path_or_s3_store(
    path_or_url: str,
    s3_kwargs: Mapping[str, Any] = None,
    s3_client_kwargs: Mapping[str, Any] = None,
    mode: str = "r",
) -> tuple[Union[str, dict], bool]:
    """If *path_or_url* is an object storage URL, return a object storage
    Zarr store (mapping object) using *s3_client_kwargs* and *mode* and
    a flag indicating whether the Zarr datasets is consolidated.

    Otherwise *path_or_url* is interpreted as a local file system path,
    returned as-is plus a flag indicating whether the Zarr datasets
    is consolidated.

    Args:
        path_or_url: A path or a URL.
        s3_kwargs: keyword arguments for S3 file system.
        s3_client_kwargs: keyword arguments for S3 boto3 client.
        mode: access mode "r" or "w". "r" is default.

    Returns:
        A tuple (path_or_obs_store, consolidated).
    """
    if is_s3_url(path_or_url) or s3_kwargs is not None or s3_client_kwargs is not None:
        s3, root = parse_s3_fs_and_root(
            path_or_url,
            s3_kwargs=s3_kwargs,
            s3_client_kwargs=s3_client_kwargs,
            mode=mode,
        )
        consolidated = mode == "r" and s3.exists(f"{root}/.zmetadata")
        return (
            s3fs.S3Map(root=root, s3=s3, check=False, create=mode == "w"),
            consolidated,
        )
    else:
        consolidated = os.path.exists(os.path.join(path_or_url, ".zmetadata"))
        return path_or_url, consolidated


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def parse_s3_fs_and_root(
    s3_url: str,
    s3_kwargs: Mapping[str, Any] = None,
    s3_client_kwargs: Mapping[str, Any] = None,
    mode: str = "r",
) -> tuple[s3fs.S3FileSystem, str]:
    """Parses *s3_url*, *s3_kwargs*, *s3_client_kwargs* and returns a
    new tuple (*obs_fs*, *root_path*). For example::

        obs_fs, root_path = parse_s3_fs_and_root(s3_url, s3_kwargs, s3_client_kwargs)
        obs_map = s3fs.S3Map(root=root_path, s3=obs_fs)

    Args:
        s3_url: Object storage URL, e.g. "s3://bucket/root", or
            "https://bucket.s3.amazonaws.com/root".
        s3_kwargs: keyword arguments for S3 file system.
        s3_client_kwargs: keyword arguments for S3 boto3 client.
        mode: Access mode "r" or "w",  "r" is default.

    Returns:
        A tuple (*obs_fs*, *root_path*).
    """

    root, s3_kwargs, s3_client_kwargs = parse_s3_url_and_kwargs(
        s3_url, s3_kwargs=s3_kwargs, s3_client_kwargs=s3_client_kwargs
    )
    s3 = new_s3_file_system(
        s3_kwargs=s3_kwargs,
        s3_client_kwargs=s3_client_kwargs,
        check_path=root if mode == "r" else None,
    )
    return s3, root


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def new_s3_file_system(
    s3_kwargs: Mapping[str, Any] = None,
    s3_client_kwargs: Mapping[str, Any] = None,
    s3_config_param_name: str = "s3_kwargs",
    check_path: str = None,
) -> s3fs.S3FileSystem:
    """Wrapper for s3fs.S3FileSystem() constructor that issues warnings
    in case the file system can not be created.

    Args:
        s3_kwargs: keyword arguments for S3 file system.
        s3_client_kwargs: keyword arguments for S3 boto3 client.
        s3_config_param_name: the name of the configuration parameter.
        check_path: an optional root path. If given, we call
            s3.exists(check_path) as a check whether S3 file system is
            valid.

    Returns:
        A s3fs.S3FileSystem instance.
    """

    s3_kwargs = s3_kwargs or {}
    if "use_listings_cache" not in s3_kwargs:
        # The default is not to cache any directory listings
        # because we want current contents
        s3_kwargs["use_listings_cache"] = False
    client_kwargs = s3_kwargs.pop("client_kwargs", {})
    client_kwargs.update(s3_client_kwargs or {})
    try:
        s3 = s3fs.S3FileSystem(**s3_kwargs, client_kwargs=client_kwargs)
        if check_path is not None:
            # Force potential NoCredentialsError
            s3.exists(check_path)
        return s3
    except botocore.exceptions.NoCredentialsError:
        if not s3_kwargs.get("anon"):
            warnings.warn(
                "No object storage credentials were"
                " passed or found.\n"
                "If you intend to access a public object storage,"
                " please pass " + s3_config_param_name + '={"anon": True, ...}\n'
            )
        raise


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def parse_s3_url_and_kwargs(
    s3_url: str,
    s3_kwargs: Mapping[str, Any] = None,
    s3_client_kwargs: Mapping[str, Any] = None,
) -> tuple[str, dict[str, Any], dict[str, Any]]:
    """Parses *obs_url*, *s3_kwargs*, *s3_client_kwargs* and returns a
    new tuple (*root*, *s3_kwargs*, *s3_client_kwargs*) with updated kwargs whose elements
    can be passed to the s3fs.S3FileSystem and s3fs.S3Map constructors as follows:::

        obs_fs = s3fs.S3FileSystem(**s3_kwargs, client_kwargs=s3_client_kwargs)
        obs_map = s3fs.S3Map(root=root, s3=obs_fs)

    Args:
        s3_url: Object storage URL, e.g. "s3://bucket/root", or
            "https://bucket.s3.amazonaws.com/root".
        s3_kwargs: keyword arguments for S3 file system.
        s3_client_kwargs: keyword arguments for S3 boto3 client.

    Returns:
        A tuple (root, s3_kwargs, s3_client_kwargs).
    """
    endpoint_url, root = split_s3_url(s3_url)

    new_s3_client_kwargs = dict(s3_client_kwargs) if s3_client_kwargs else dict()
    if endpoint_url:
        new_s3_client_kwargs["endpoint_url"] = endpoint_url

    # The following key + secret kwargs are no longer supported in client_kwargs and are now moved into s3_kwargs
    key = secret = None
    if "provider_access_key_id" in new_s3_client_kwargs:
        key = new_s3_client_kwargs.pop("provider_access_key_id")
    if "aws_access_key_id" in new_s3_client_kwargs:
        key = new_s3_client_kwargs.pop("aws_access_key_id")
    if "provider_secret_access_key" in new_s3_client_kwargs:
        secret = new_s3_client_kwargs.pop("provider_secret_access_key")
    if "aws_secret_access_key" in new_s3_client_kwargs:
        secret = new_s3_client_kwargs.pop("aws_secret_access_key")

    new_s3_kwargs = dict(key=key, secret=secret)
    if s3_kwargs:
        new_s3_kwargs.update(**s3_kwargs)

    return root, new_s3_kwargs, new_s3_client_kwargs


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def split_s3_url(path: str) -> tuple[Optional[str], str]:
    """If *path* is a URL, return tuple (endpoint_url, root), otherwise (None, *path*)"""
    url = urllib3.util.parse_url(path)
    if all((url.scheme, url.host, url.path)) and url.scheme != "s3":
        if url.port is not None:
            endpoint_url = f"{url.scheme}://{url.host}:{url.port}"
        else:
            endpoint_url = f"{url.scheme}://{url.host}"
        root = url.path
        if root.startswith("/"):
            root = root[1:]
        return endpoint_url, root
    return None, path


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def is_s3_url(path_or_url: str) -> bool:
    """Test if *path_or_url* is a potential object storage URL.

    Args:
        path_or_url: Path or URL to test.

    Returns:
        True, if so.
    """
    return (
        path_or_url.startswith("https://")
        or path_or_url.startswith("http://")
        or path_or_url.startswith("s3://")
    )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import sys

import click


# noinspection PyShadowingBuiltins
@click.command(name="extract")
@click.argument("cube")
@click.argument("points")
@click.option(
    "--output",
    "-o",
    metavar="OUTPUT",
    help="Output path. If omitted, output is written to stdout.",
)
@click.option(
    "--format",
    "-f",
    metavar="FORMAT",
    type=click.Choice(["csv", "json", "xlsx"]),
    help="Output format. Currently, only 'csv' is supported.",
    default="csv",
)
@click.option(
    "--coords", "-C", is_flag=True, help="Include cube cell coordinates in output."
)
@click.option(
    "--bounds",
    "-B",
    is_flag=True,
    help="Include cube cell coordinate boundaries (if any) in output.",
)
@click.option(
    "--indexes", "-I", is_flag=True, help="Include cube cell indexes in output."
)
@click.option(
    "--refs", "-R", is_flag=True, help="Include point values as reference in output."
)
def extract(
    cube,
    points,
    output=None,
    format=None,
    coords=False,
    bounds=False,
    indexes=False,
    refs=False,
):
    """Extract cube points.

    Extracts data cells from CUBE at coordinates given in each POINTS record and writes the resulting values to given
    output path and format.

    POINTS must be a CSV file that provides at least the columns "lon", "lat", and "time". The "lon" and "lat"
    columns provide a point's location in decimal degrees. The "time" column provides a point's date or
    date-time. Its format should preferably be ISO, but other formats may work as well.
    """
    if format != "csv":
        raise click.ClickException(f"Format {format!r} is not supported.")

    import pandas as pd

    cube_path = cube
    points_path = points
    output_path = output
    include_coords = coords
    include_bounds = bounds
    include_indexes = indexes
    include_refs = refs

    from xcube.core.dsio import open_dataset
    from xcube.core.extract import (
        DEFAULT_INDEX_NAME_PATTERN,
        DEFAULT_REF_NAME_PATTERN,
        get_cube_values_for_points,
    )

    # We may make the following CLI options
    index_name_pattern = DEFAULT_INDEX_NAME_PATTERN
    ref_name_pattern = DEFAULT_REF_NAME_PATTERN
    time_col_names = ["time"]

    points = pd.read_csv(points_path, parse_dates=time_col_names)
    points.time = points.time.dt.tz_localize(tz=None)
    with open_dataset(cube_path) as cube:
        values = get_cube_values_for_points(
            cube,
            points,
            include_coords=include_coords,
            include_bounds=include_bounds,
            include_indexes=include_indexes,
            index_name_pattern=index_name_pattern,
            include_refs=include_refs,
            ref_name_pattern=ref_name_pattern,
        ).to_dataframe()
        values.to_csv(
            output_path if output_path else sys.stdout,
            # We may make the following CLI options
            sep=",",
            date_format="%Y-%m-%dT%H:%M:%SZ",
            index=True,
        )
